/*
	(C) 2001 AROS - The Amiga Research OS
	$Id: main.c,v 1.41 2012/04/01 19:43:41 jacadcaps Exp $
 
	Desc:
	Lang: English
*/


/*********************************************************************************************/

#define USE_NEWSTARTUPSYNC			1

/*********************************************************************************************/

#define AROS_ALMOST_COMPATIBLE
#include "global.h"
#include "sample.h"
#include <exec/execbase.h>
#include <exec/system.h>
#include <clib/alib_protos.h>
#include <proto/input.h>
#include <private/vapor/vapor.h>
#include <proto/mount.h>

#define DEBUG 0
#include <aros/debug.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifndef MUIA_Application_NoIconify
#define MUIA_Application_NoIconify          0x80426a3b
#endif

#ifndef MUIA_Application_UseScreenNotify
#define MUIA_Application_UseScreenNotify    0x80420861 
#endif

/*********************************************************************************************/
struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;
struct IntuitionBase	*IntuitionBase;
struct GfxBase		*GfxBase;
struct Library		*IFFParseBase;
struct Library		*UtilityBase;
struct Library		*LocaleBase;
struct Library		*KeymapBase;
struct Library		*LayersBase;
struct Library		*DataTypesBase;
struct Library		*DiskfontBase;
struct Library		*WorkbenchBase;
struct Library		*IconBase;
struct Library		*CyberGfxBase;
struct Library      *QueryBase;
struct Library		*MountBase;

static struct libinfo
{
	APTR        var;
	STRPTR      name;
	WORD        version;
	BOOL        nocloseafterpatch;
}
libtable[] =
{
	{&IntuitionBase     , "intuition.library"           , 39, FALSE },
	{&GfxBase           , "graphics.library"            , 39, FALSE },
	{&UtilityBase       , "utility.library"             , 39, TRUE  },
	{&IFFParseBase      , "iffparse.library"            , 39, FALSE },
	//    {&LocaleBase        , "locale.library"              , 39, TRUE  },
	{&KeymapBase        , "keymap.library"              , 39, FALSE },
	{&LayersBase        , "layers.library"              , 39, FALSE },
	//    {&DataTypesBase     , "datatypes.library"           , 39, FALSE },
	{&DiskfontBase      , "diskfont.library"            , 39, TRUE  },
	{&WorkbenchBase     , "workbench.library"           , 39, TRUE  },
	{&IconBase          , "icon.library"                , 39, TRUE  },
	{&CyberGfxBase      , "cybergraphics.library"       , 50, TRUE  },
	{&QueryBase         , "query.library"               , 51, TRUE  },
	{&MountBase         , "mount.library"               , 51, TRUE  },
	{NULL   	    	    	    	    	    	    	    }
};
/*********************************************************************************************/

static struct prefinfo
{
	STRPTR                  filename;
	STRPTR                  filenamebuffer;
	BOOL                    nextisfallback;
	ULONG			(*handler)(char *filename);
	struct NotifyRequest    nr;
	BOOL                    notifystarted;
}

preftable[] =
{
	{"mouse.conf"		, inputmousetxtprefsname,	TRUE, (ULONG (*)(char*))LoadMousePrefs},
	{"mouse.prefs"		, inputmouseprefsname,		FALSE, NULL},

	{"keyboard.conf"	, inputkeyboardtxtprefsname,	TRUE, (ULONG (*)(char *)) LoadKeyboardPrefs },
	{"keyboard.prefs"	, inputkeyboardprefsname,	FALSE, NULL},

	{"input.prefs"		, inputprefsname,		FALSE, NULL},

	{"font.conf"		, newfontprefsname,		TRUE, (ULONG (*)(char *)) LoadFontPrefs },
	{"font.prefs"		, fontprefsname,		FALSE, NULL},
	
	{"/mui/screens.txt"	 , screensname,		   TRUE, (ULONG (*)(char *)) LoadScreenPrefs },
	{"publicscreens.prefs"	, publicscreensname,		FALSE, NULL },
	//{"screenmode.prefs" , screenprefsname,	  FALSE, NULL},
	
	{"locale.prefs"		, localeprefsname,		FALSE, NULL},
	
	{"palette.prefs"	, paletteprefsname,		FALSE, NULL},
	
	{"wbpattern.prefs"	, patternprefsname,		FALSE, NULL},
	
	{"icontrol.conf"	, icontroltxtprefsname,		TRUE, (ULONG (*)(char *)) LoadIControl },
	{"newicontrol.prefs"	, newicontrolprefsname,		FALSE, NULL},
	//{"icontrol.prefs"	  , icontrolprefsname,		  FALSE, NULL},
	
	{"serial.prefs"		, serialprefsname,		FALSE, NULL},
	
	{"printer.prefs"	, printerprefsname,		FALSE, NULL},
	{"printergfx.prefs"	, printergfxprefsname,		FALSE, NULL},
	
	{"pointer.prefs"	, pointerprefsname,		FALSE, NULL},
	
	{"overscan.prefs"	, overscanprefsname,		FALSE, NULL},
	
	{"sound.prefs"		, soundprefsname,		FALSE, NULL},
	
	/* WB 3.5+ */
	{"workbench.prefs"	, workbenchprefsname,		FALSE, NULL},
	{"reaction.prefs"	, reactionprefsname,		FALSE, NULL},
	
	/* POSEIDON */
	{"poseidon.prefs"	, poseidonprefsname,		FALSE, (ULONG (*)(char *)) LoadPoseidonPrefs },

	/* screensaver */
	{"blanker.conf"		, newblankerprefsname,		TRUE, (ULONG (*)(char *)) LoadBlankerPrefs },
	{"blanker.prefs"	, blankerprefsname,		FALSE, NULL},
	
	/* network */
	{"interface.prefs"	, networkprefsname,		FALSE, (ULONG (*)(char *)) LoadNetworkPrefs },
	{"user.prefs"		, userprefsname,		FALSE, (ULONG (*)(char *)) LoadUserPrefs },
	
	/* log */
	{"log.prefs"		, logserverprefsname,		FALSE, (ULONG (*)(char *)) LoadLogServerPrefs },

	/* compositing */
	{"displayengine.conf" , compositingprefsname,		 FALSE, (ULONG (*)(char *)) LoadCompositingPrefs },
	
	/* powerbook */
	{"laptop.conf", powerbookprefsname, FALSE, (ULONG (*)(char *)) LoadPowerBookPrefs },
	{"laptopquick.conf", powerbook2prefsname, FALSE, (ULONG (*)(char *)) LoadPowerBookPrefs },

	/* end */
	{NULL}
};

static STRPTR GenerateKeyboards[] =
{
	"pc104_pl",
	"pc104_cz",
	"pc104_cz_qwerty",
	"pc105_hu",
	"pc105_hu_qwerty",
	NULL
};

struct LanguageFontDescr
{
	char *Language;
	char *FontFile;
};

static struct LanguageFontDescr FontLanguages[] =
{
	{"polski","Type=1 FontName=\"Lux Mono.font\" FontYSize=13 FontStyle=0x0 FontFlags=0x0\nType=2 FontName=\"Lux.font\" FontYSize=15 FontStyle=0x0 FontFlags=0x0\n"},
	{"magyar","Type=1 FontName=\"Lux Mono.font\" FontYSize=13 FontStyle=0x0 FontFlags=0x0\nType=2 FontName=\"Lux.font\" FontYSize=15 FontStyle=0x0 FontFlags=0x0\n"},
	{"etina","Type=1 FontName=\"DejaVu Sans Mono.font\" FontYSize=14 FontStyle=0x0 FontFlags=0x0\nType=2 FontName=\"DejaVu Sans.font\" FontYSize=16 FontStyle=0x0 FontFlags=0x0\n"},
	{NULL,NULL}
};

/*********************************************************************************************/

static struct scaninfo
{
	ULONG       id;
	struct Hook hook;
}
scantable[] =
	{
		{ ID_INPT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(InputPrefs_Handler) }}, /* obsolete */
		{ ID_SCRM, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(ScreenPrefs_Handler) }},
		{ ID_FONT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(FontPrefs_Handler) }},
		{ ID_LCLE, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(LocalePrefs_Handler) }},
		{ ID_PALT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(PalettePrefs_Handler) }},
		{ ID_PNTR, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(PointerPrefs_Handler) }},
		{ ID_PTRN, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(WBPatternPrefs_Handler) }},
		{ ID_SOND, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(SoundPrefs_Handler) }},
		{ ID_ICTL, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(IControlPrefs_Handler) }},
		{ ID_PGFX, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(PrinterGfxPrefs_Handler) }},
		{ ID_PTXT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(PrinterTxtPrefs_Handler) }},
		{ ID_PUNT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(PrinterUnitPrefs_Handler) }},
		/* WB 3.5+ */
		{ ID_WBNC, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(WorkbenchPrefs_Handler) }},
		{ ID_WBHD, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(WorkbenchHiddenDevicePrefs_Handler) }},
		{ ID_RACT, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(ReactionPrefs_Handler) }},
		/* V50 */
		{ ID_ICTA, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(IControlActions_Handler) }},
		{ ID_ICTH, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(IControlHotkeys_Handler) }},
		{ ID_ICTE, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(IControlExtensions_Handler) }},
		{ ID_INPE, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(InputPrefsExt_Handler) }},
		{ ID_INPM, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(InputMousePrefs_Handler) }},
		{ ID_INPK, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(InputKeyboardPrefs_Handler) }},
		{ ID_SERL, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(SerialPrefs_Handler) }},
		{ MAKE_ID('M','P','U','B'), {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(MUIScreenPrefs_Handler) }},
		{ ID_BLNK, {{NULL, NULL}, (APTR)AROS_ASMSYMNAME(BlankerPrefs_Handler) }},
	};


/*********************************************************************************************/

static ULONG notifysig;
static ULONG reopenwbtimersig;

/*********************************************************************************************/

#if USE_DELAYEDREOPENWB
void		DoReOpenWBScreen(void);
#endif /* USE_DELAYEDREOPENWB */

static void	InitAC97Mixer(void);
static ULONG	OpenLibs(void);
static void	CloseLibs(void);
static void	PrintMessage(LONG logo, STRPTR msg);
static void	PrintMessageReq(LONG logo, STRPTR msg);
static ULONG	StartNotifications(void);
static void	KillNotifications(void);
static void	IPrefsHandler(void);
static ULONG	GetENVName(void);
static ULONG	FixENVDirs(CONST_STRPTR base);
static ULONG	PreparePatches(void);
static void	ValidateInstallation(void);
static void SetBootCDLanguageKB(void);
static void DumpFonts(void);
static struct Message gmsg;
static struct MsgPort *grport;

/*********************************************************************************************/

int main(void)
{
	int ret = RETURN_FAIL;
	SysBase=*((struct ExecBase**) 4);

	D(bug("IPrefs started\n"));

	if (FindTask(IPREFS_SEM_NAME))
	{
		D(bug("IPrefs already running, quit!\n"));
		return ret;
	}

	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37)))
	{
		ULONG OldSegList = NULL;
		LocaleBase = OpenLibrary("locale.library", 37);

		if (OpenLibs())
		{
			if (MountBase->lib_Version >= 52 && MountBase->lib_Revision >= 3)
				MountMountDiskHandlersTags(TAG_DONE);

			if (GetENVName())
			{
				/* Create possibly missing directories */
				FixENVDirs(envname);
				FixENVDirs("ENVARC:");

				if ((grport = CreateMsgPort()))
				{
					struct Process	*MyProcess;
					struct CommandLineInterface *cli;

					gmsg.mn_Node.ln_Type = NT_REPLYMSG;
					gmsg.mn_ReplyPort    = NULL;
					gmsg.mn_Length       = sizeof(gmsg);

					MyProcess = CreateNewProcTags(NP_Entry,(ULONG) &IPrefsHandler,
					                              NP_Name,(ULONG) IPREFS_SEM_NAME,
					                              NP_Priority,0,
					                              NP_CodeType,MACHINE_PPC,
					                              NP_ConsoleTask, NULL,
					                              NP_CurrentDir, 0,
					                              NP_CopyVars, FALSE,
					                              NP_WindowPtr, NULL,
					                              NP_HomeDir, 0,
					                              TAG_END);
					if (MyProcess)
					{
						D(bug("IPrefs: Process 0x%lx\n",
						      (ULONG)MyProcess));
						cli			=	Cli();
						OldSegList		=	cli->cli_Module;
						cli->cli_Module	=	NULL;

						/*
						 * We must wait until the prefs loading is done.
						 */
						WaitPort(grport);
						(void) GetMsg(grport);

						ValidateInstallation();

						ret = RETURN_OK;
					}
					else
					{
						D(bug("IPrefs: CreateNewProc failed\n"));
					}
					DeleteMsgPort(grport);
				}
				else
				{
					D(bug("IPrefs: No MsgPort\n"));
				}
			}
			else
			{
				PrintMessage(-1, "Error expanding \"ENV:\" to full name!");
			}
		}
		if (!OldSegList)
		{
			CloseLibrary(LocaleBase);
			CloseLibs();
			CloseLibrary((struct Library *)DOSBase);
		}
	}

	return ret;
}

/*********************************************************************************************/

static ULONG ReadQualifier(void)
{
	struct IORequest	MyIORequest;
	struct Library		*InputBase;
	ULONG			Qualifier=0;

	memset(&MyIORequest, 0, sizeof(MyIORequest));
	if (OpenDevice("input.device",
	               0,
	               &MyIORequest,
	               0)==0)
	{
		InputBase	=(struct Library*) MyIORequest.io_Device;
		Qualifier	=	PeekQualifier();
		CloseDevice(&MyIORequest);
	}

	return(Qualifier);
}

/* This routine checks and reports for any problems in the installation.
*/
static void ValidateInstallation(void)
{
	BPTR mossys;
	UBYTE _id[sizeof(struct InfoData) + 3];
	struct InfoData *id = (APTR) ((((ULONG) _id) + 3) & ~3);
	struct Process *me = (struct Process *) FindTask(NULL);
	APTR oldwinptr = me->pr_WindowPtr;
	struct Resident *morphosres;

#if 1
	morphosres = FindResident("MorphOS");

	if (morphosres == 0
	|| (morphosres->rt_Flags&RTF_EXTENDED) == 0
	|| morphosres->rt_Version < MORPHOSVERSION
#if MORPHOSREVISION > 0
	|| (morphosres->rt_Version == MORPHOSVERSION && morphosres->rt_Revision < MORPHOSREVISION )
#endif
	)
	{
		BOOL cdsystem = TRUE;

		if (SysBase->MaxLocMem)
		{
			/* It's amiga, don't whine about OF related cd booting */
			cdsystem = FALSE;
		}
		else /* pegasos, efika, something else */
		{
			UBYTE SystemName[80];
			LONG len;

			len = NewGetSystemAttrs(SystemName, sizeof(SystemName) - 1, SYSTEMINFOTYPE_SYSTEM, TAG_DONE);
			if (len < 0) len = 0;
			SystemName[len] = '\0';

			/* If it's Efika, don't whine about CD-ROM */
			if (!strcmp(SystemName, "bplan,EFIKA5K2"))
			{
				cdsystem = FALSE;
			}
		}

		if (cdsystem)
		{
			BOOL cdboot = FALSE;

			me->pr_WindowPtr = (APTR) -1;

			if (Info(0, id) && id->id_DiskState == ID_WRITE_PROTECTED &&
			    id->id_DiskType == ID_CDROM_ISO_DISK)
			{
				BPTR lock;

				lock = Lock("SYS:", ACCESS_READ);
				if (lock)
				{
					UBYTE name[32];

					if (NameFromLock(lock, name, sizeof(name)) &&
					   !strcmp(name, "MorphOSBoot:"))
					{
						cdboot = TRUE;
					}

					UnLock(lock);
				}
			}

			me->pr_WindowPtr = oldwinptr;

			if (cdboot)
			{
				PrintMessageReq(-1, "MorphOS boot.img version is too old.\n"
				                "\n"
				                "In order to boot from the MorphOS CD, type\n"
				                "boot /pci/cd\n"
				                "to the OpenFirmware prompt."
				                );
			}
			else
			{
				PrintMessageReq(-1, "MorphOS boot.img version is too old.\n"
				                "\n"
				                "You forgot to update boot.img of the boot partition.\n"
				                "You are advised to boot from the MorphOS CD and to\n"
				                "use the HDInstall program.\n"
				                "\n"
				                "In order to boot from the MorphOS CD, type\n"
				                "boot /pci/cd\n"
				                "to the OpenFirmware prompt."
				                );
			}
		}
		else
		{
			/* Generic complaint message for systems without OF or CD-ROM possibility */

			PrintMessageReq(-1, "MorphOS boot.img version is too old.\n"
			                "\n"
			                "Make sure you're using the correct\n"
			                "boot.img file."
			                );
		}

		PrintMessageReq(-1, "System will reboot now.");

		if ((ReadQualifier() & (IEQUALIFIER_LSHIFT|IEQUALIFIER_RSHIFT)) == 0)
		{
			ColdReboot();
		}
		else
		{
			SystemTags("newshell syscon:", SYS_Input, 0, SYS_Output, 0, SYS_Asynch, TRUE, TAG_DONE);
			Wait(0);
		}
	}

#endif

	if (DOSBase->dl_Root->rn_Flags & RNF_SYS_CONFLICT)
	{
		DOSBase->dl_Root->rn_Flags &= (ULONG) ~RNF_SYS_CONFLICT;

		PrintMessageReq(-1, "Volume/Device SYS: exists.\n"
		                "\n"
		                "SYS name is reserved for system, and\n"
		                "such Volume/Device cannot exist.\n"
		                "\n"
		                "This Volume/Device has been temporarily\n"
		                "renamed. You should fix the conflict by\n"
		                "relabeling the Volume/Device in question.");
	}

	if (DOSBase->dl_Root->rn_Flags & RNF_MOSSYS_CONFLICT)
	{
		DOSBase->dl_Root->rn_Flags &= (ULONG) ~RNF_MOSSYS_CONFLICT;

		PrintMessageReq(-1, "Volume/Device MOSSYS: exists.\n"
		                "\n"
		                "MOSSYS name is reserved for system, and\n"
		                "such Volume/Device cannot exist.\n"
		                "\n"
		                "This Volume/Device has been temporarily\n"
		                "renamed. You should fix the conflict by\n"
		                "relabeling the Volume/Device in question.");
	}

	me->pr_WindowPtr = (APTR) -1;

	/* get MOSSYS: lock */
	mossys = Lock("MOSSYS:", ACCESS_READ);

	/* Only if disk is writeable
	 */
	if (mossys && Info(0, id) && id->id_DiskState == ID_VALIDATED)
	{
		BPTR ramlock;

		ramlock = Lock("RAM:", ACCESS_READ);
		if (ramlock)
		{
			CONST_STRPTR assigns[] =
			{
				"L",
				"Fonts",
				"Devs",
				"Libs",
				"S",
				"C",
				"Locale",
				NULL
			};
			CONST_STRPTR *assignpt = assigns;
			CONST_STRPTR name;
			BPTR olddir;
			UBYTE msg[256];
			LONG miss = 0;

			/* The boot disk.. */
			olddir = CurrentDir(0);

			strcpy(msg, "Following directories were missing from the\n"
			            "boot disk:\n\n");

			while ((name = *assignpt++))
			{
				BPTR lock;

				lock = Lock(name, ACCESS_READ);
				if (!lock && IoErr() == ERROR_OBJECT_NOT_FOUND)
				{
					int ok = FALSE;

					strcat(msg, name);

					lock = CreateDir(name);
					if (lock)
					{
						/* Free the exclusive dirlock */
						UnLock(lock);

						lock = Lock(name, ACCESS_READ);
						if (lock)
						{
							/* Find the assign on RAM: and replace with
							 * the newly created directory.
							 */
							struct DosList *dlist, *dl;

							dlist = LockDosList(LDF_ASSIGNS | LDF_WRITE);
							dl = FindDosEntry(dlist, name, LDF_ASSIGNS);

							if (dl && dl->dol_Type != DLT_LATE &&
							    dl->dol_Type != DLT_NONBINDING)
							{
								if (SameLock(dl->dol_Lock, ramlock) == LOCK_SAME_VOLUME)
								{
									BPTR oldlock;

									/* Swap in the new lock - a bit magic
									 */
									oldlock = dl->dol_Lock;
									dl->dol_Lock = lock;
									lock = oldlock;
								}
							}
							UnLockDosList(LDF_ASSIGNS | LDF_WRITE);

							UnLock(lock);
						}

						ok = TRUE;
					}

					if (ok)
						strcat(msg, " - create ok\n");
					else
						strcat(msg, " - create failed\n");

					miss++;
				}
				else
					UnLock(lock);
			}

			if (miss)
			{
				strcat(msg, "\n"
				            "These directories were created for your\n"
				            "convenience.");

				me->pr_WindowPtr = oldwinptr;

				PrintMessageReq(-1, msg);
			}

			CurrentDir(olddir);

			UnLock(ramlock);
		}
	}

	UnLock(mossys);

	me->pr_WindowPtr = oldwinptr;
}

/*********************************************************************************************/

struct KeymapDescription
{
	struct MinNode node;
	char           keymapname[64];
	char           dispname[128];
};

struct LanguageDescription
{
	struct MinNode node;
	char           dispname[128];
};

/* ripped off kemap.mprefs */
static struct Resident *FindLibResident(BPTR MySegList)
{
  struct Resident	*MyResident;
  ULONG			*MySegment;
  UWORD			*MyBuffer;
  UWORD			*EndBuffer;

  while (MySegList)
  {
#define RESIDENT_SIZEOF 26
	MySegment	=(ULONG*) BADDR(MySegList);
	MyBuffer	=(UWORD*) &MySegment[1];
	EndBuffer	=(UWORD*) (((UBYTE *) MySegment) + MySegment[-1] - RESIDENT_SIZEOF - sizeof(ULONG));

	while (MyBuffer < EndBuffer)
	{
	  if (*MyBuffer == RTC_MATCHWORD)
	  {
		MyResident=(struct Resident*) MyBuffer;
		if (MyResident->rt_MatchTag == MyResident)
		{
			return MyResident;
		}
		else
		{
		}
	  }
	  MyBuffer++;
	}
	MySegList	=(BPTR) MySegment[0];
  }
  return NULL;
}

static void AddEntry(char *Name,struct List *keymaps,APTR pool)
{
	BPTR MySegList;

	// load the segment and examine it

	if ((MySegList = LoadSeg(Name)))
	{
		struct Resident *MyResident;

		// check its resident
		if ((MyResident=FindLibResident(MySegList)))
		{
			char *description = NULL;

			// we only support new keymaps here, but it doesn't matter much since
			// only new mode keymaps are on the iso
			if (QueryGetResidentAttr(MyResident,&description,QUERYINFOATTR_DESCRIPTION) && description && *description)
			{
				struct KeymapDescription *us = AllocPooled(pool,sizeof (struct KeymapDescription));

				if (us)
				{
					stccpy(us->keymapname,Name,sizeof(us->keymapname));
					stccpy(us->dispname,description,sizeof(us->dispname));

					ADDTAIL(keymaps,us);
				}
			}
		}

		UnLoadSeg(MySegList);
	}
}

static void MakeKeyMapsList(struct List *keymaps,APTR pool)
{
	struct AnchorPath *AnchorPath;

	// scan the mossys:devs/keymaps dir
	if ((AnchorPath = (struct AnchorPath *) AllocVec(sizeof(struct AnchorPath), MEMF_PUBLIC|MEMF_CLEAR)))
	{
		LONG ret;
		BPTR moslock;

		AnchorPath->ap_Flags     = 0;
		AnchorPath->ap_BreakBits = 0;
		AnchorPath->ap_Strlen    = 0;

		moslock = Lock("MOSSYS:Devs/Keymaps/", ACCESS_READ);

		if (moslock)
		{
			BPTR olddir;

			olddir = CurrentDir(moslock);

			ret = MatchFirst("~(#?.info)", AnchorPath);

			while (!ret)
			{
				if (AnchorPath->ap_Info.fib_DirEntryType < 0)
				{
					// try adding each entry
					AddEntry(AnchorPath->ap_Info.fib_FileName,keymaps,pool);
				}

				ret = MatchNext(AnchorPath);
			}
			MatchEnd(AnchorPath);

			CurrentDir(olddir);
			UnLock(moslock);

		}

		FreeVec((APTR)AnchorPath);
	}
}

/* ripped off locale.mprefs */
static void MakeLanguagesList(struct List *languages,APTR pool)
{
	struct AnchorPath *AnchorPath;

	if ((AnchorPath = (struct AnchorPath *)AllocVec(sizeof(struct AnchorPath), MEMF_PUBLIC)))
	{
		struct Process *Self = (struct Process *) FindTask(NULL);
		APTR oldwinptr;
		LONG ret;

		oldwinptr = Self->pr_WindowPtr;
		Self->pr_WindowPtr = (APTR) -1;

		AnchorPath->ap_Flags     = 0;
		AnchorPath->ap_BreakBits = 0;
		AnchorPath->ap_Strlen    = 0;

		ret = MatchFirst("MOSSYS:locale/languages/#?.language", AnchorPath);

		while (!ret)
		{
			if (AnchorPath->ap_Info.fib_DirEntryType < 0)
			{
				struct LanguageDescription *us = AllocPooled(pool,sizeof (struct LanguageDescription));
				char *point;
				char name[109];
				int k;

				point = strpbrk(AnchorPath->ap_Info.fib_FileName, ".");
				stccpy(name, AnchorPath->ap_Info.fib_FileName, point - AnchorPath->ap_Info.fib_FileName + 1);

				name[0] = ToUpper(name[0]);
				if (name[0])
				{
					for (k = 1; name[k]; k++)
						name[k] = ToLower(name[k]);
				}

				if (us)
				{
					stccpy(us->dispname,name,sizeof(us->dispname));

					ADDTAIL(languages,us);
				}

			}
			ret = MatchNext(AnchorPath);
		}

		MatchEnd(AnchorPath);

		Self->pr_WindowPtr = oldwinptr;

		FreeVec((APTR)AnchorPath);
	}
}

MUI_HOOK(KeymapDispl, char **text, struct KeymapDescription *entry)
{
	text[0] = entry ? entry->dispname : "Keymap";
	return 0;
}

MUI_HOOK(KeymapCompare, struct KeymapDescription *entry1, struct KeymapDescription *entry2)
{
	return Stricmp(entry2->dispname,entry1->dispname);
}

MUI_HOOK(LangDispl, char **text, struct LanguageDescription *entry)
{
	text[0] = entry ? entry->dispname : "Preferred Language";
	return 0;
}

MUI_HOOK(LangCompare, struct LanguageDescription *entry1, struct LanguageDescription *entry2)
{
	return Stricmp(entry2->dispname,entry1->dispname);
}

static void SetBootCDLanguageKB(void)
{
	UBYTE _id[sizeof(struct InfoData) + 3];
	struct InfoData *id = (APTR) ((((ULONG) _id) + 3) & ~3);
	struct Process *me = (struct Process *) FindTask(NULL);
	APTR oldwinptr = me->pr_WindowPtr;
	BOOL cdboot = FALSE;

	me->pr_WindowPtr = (APTR) -1;

	if (Info(0, id) && id->id_DiskState == ID_WRITE_PROTECTED &&
	    id->id_DiskType == ID_CDROM_ISO_DISK)
	{
		BPTR lock;

		lock = Lock("SYS:", ACCESS_READ);

		if (lock)
		{
			UBYTE name[32];

			if (NameFromLock(lock, name, sizeof(name)) && !strcmp(name, "MorphOSBoot:"))
			{
				cdboot = TRUE;
			}

			UnLock(lock);
		}
	}

	me->pr_WindowPtr = oldwinptr;

	if (cdboot)
	{
		APTR pool = CreatePool(MEMF_PUBLIC,1024,1024);

		if (pool)
		{
			struct List keymaps;
			struct List languages;

			NEWLIST(&keymaps);
			NEWLIST(&languages);

			MakeKeyMapsList(&keymaps,pool);
			MakeLanguagesList(&languages,pool);

			if (!IsListEmpty(&keymaps))
			{
				struct Library *MUIMasterBase = OpenLibrary("muimaster.library",20);

				if (MUIMasterBase)
				{
					Object *app,*win;
					Object *keymaplist,*languageslist;
					Object *button;

					app = MUI_NewObject(MUIC_Application,
						MUIA_Application_Title, (ULONG)"MorphOS Boot",
						MUIA_Application_Copyright, (ULONG)" 2012 MorphOS Team",
						MUIA_Application_Description, (ULONG)"Pre-boot language/keymap selector",
						MUIA_Application_Author, (ULONG)"Jacek Piszczek",
						MUIA_Application_NoIconify, TRUE,
						MUIA_Application_UseScreenNotify, FALSE,

						SubWindow, (ULONG)win = MUI_NewObject(MUIC_Window,
							MUIA_Window_Title, (ULONG)"MorphOS Boot",
							MUIA_Window_CloseGadget, FALSE,
							MUIA_Window_Width, 500,
							MUIA_Window_Height, 800,
							WindowContents, (ULONG)MUI_NewObject(MUIC_Group,MUIA_Group_Horiz, FALSE,
								Child, (ULONG)languageslist = MUI_NewObject(MUIC_List,
									MUIA_List_Title, TRUE,
									InputListFrame,
									MUIA_List_DisplayHook, (ULONG)&LangDispl_hook,
									MUIA_List_CompareHook, (ULONG)&LangCompare_hook,
									TAG_DONE),
								Child, (ULONG)keymaplist = MUI_NewObject(MUIC_List,
									MUIA_List_Title, TRUE,
									InputListFrame,
									MUIA_List_DisplayHook, (ULONG)&KeymapDispl_hook,
									MUIA_List_CompareHook, (ULONG)&KeymapCompare_hook,
									TAG_DONE),
								Child, (ULONG)button = MUI_MakeObject(MUIO_Button,(ULONG)"OK"),
								TAG_DONE),
							TAG_DONE),
						TAG_DONE);

					if (app)
					{
						struct KeymapDescription *kd;
						struct KeymapDescription usa1kd;
						struct LanguageDescription *ld;
						struct LanguageDescription engld;

						strcpy(usa1kd.keymapname,"usa1");
						strcpy(usa1kd.dispname,"American");
						strcpy(engld.dispname,"English");

						// ui setup
						set(button,MUIA_CycleChain,TRUE);
						set(languageslist,MUIA_CycleChain,TRUE);
						set(keymaplist,MUIA_CycleChain,TRUE);
						DoMethod(button,MUIM_Notify,MUIA_Pressed,FALSE,(ULONG)app,2,MUIM_Application_ReturnID,MUIV_Application_ReturnID_Quit);

						// populate lists
						ForeachNode(&keymaps,(struct Node *)kd)
						{
							DoMethod(keymaplist,MUIM_List_InsertSingle,(ULONG)kd,MUIV_List_Insert_Bottom);
						}
						DoMethod(keymaplist,MUIM_List_Sort);
						DoMethod(keymaplist,MUIM_List_InsertSingle,(ULONG)&usa1kd,MUIV_List_Insert_Top);
						set(keymaplist,MUIA_List_Active,0);

						ForeachNode(&languages,(struct Node *)kd)
						{
							DoMethod(languageslist,MUIM_List_InsertSingle,(ULONG)kd,MUIV_List_Insert_Bottom);
						}
						DoMethod(languageslist,MUIM_List_Sort);
						DoMethod(languageslist,MUIM_List_InsertSingle,(ULONG)&engld,MUIV_List_Insert_Top);
						set(languageslist,MUIA_List_Active,0);

						// open the window
						set(win,MUIA_Window_ActiveObject,(ULONG)button);
						set(win,MUIA_Window_Open,TRUE);

						// standard loop, quits on OK
						{
							ULONG sigs = 0;

							while (DoMethod(app,MUIM_Application_NewInput,(ULONG)&sigs) != MUIV_Application_ReturnID_Quit)
							{
								if (sigs)
								{
									sigs = Wait(sigs | SIGBREAKF_CTRL_C);
									if (sigs & SIGBREAKF_CTRL_C) break;
								}
							}
						}

						// close window
						set(win,MUIA_Window_Open,FALSE);

						// get the selection
						kd = NULL;
						ld = NULL;

						DoMethod(keymaplist,MUIM_List_GetEntry,MUIV_List_GetEntry_Active,(ULONG)&kd);
						DoMethod(languageslist,MUIM_List_GetEntry,MUIV_List_GetEntry_Active,(ULONG)&ld);

						// valid selection and not the default kb?
						if ((kd != NULL) && (kd != &usa1kd))
						{
							// save new conf file!
							BPTR f;
							char sbuffer[512];
							char *extra = "";
							int i = 0;

							while (GenerateKeyboards[i])
							{
								if (strcmp(GenerateKeyboards[i],kd->keymapname) == 0) extra = "GenerateFT2Map";
								i++;
							}

							sprintf(sbuffer,"Keyboard=\"%s\" %s\n",kd->keymapname,extra);

							if ((f = Open("ENV:sys/keyboard.conf",MODE_NEWFILE)) != NULL)
							{
								Write(f,sbuffer,strlen(sbuffer));
								Close(f);
							}
						}

						D(bug("Keymap changed...\n"));

						// language changed?
						if ((ld != NULL) && (ld != &engld))
						{
							struct IFFHandle *ip;
							struct LocalePrefs loc;
							struct Locale *def = OpenLocale(NULL);

							memset(&loc,0,sizeof (loc));

							if (def != NULL)
							{
								int i = 0;
								// fill in most values from the default locales.. we just want to slightly tweak the config
								strcpy(loc.lp_CountryName,"united_states");
								stccpy(loc.lp_PreferredLanguages[0],ld->dispname,sizeof(loc.lp_PreferredLanguages[0]));
								loc.lp_PreferredLanguages[0][0] |= 0x20;
								strcpy(loc.lp_PreferredLanguages[1],"english"); // always second!
								loc.lp_GMTOffset = def->loc_GMTOffset;
								loc.lp_Flags = def->loc_Flags;
								loc.lp_CountryData.cp_CountryCode = def->loc_CountryCode;
								loc.lp_CountryData.cp_TelephoneCode = def->loc_TelephoneCode; //eeeks!
								loc.lp_CountryData.cp_MeasuringSystem = def->loc_MeasuringSystem;

								#define LOCALESTRCOPY(what) \
									stccpy(loc.lp_CountryData.cp_##what,def->loc_##what,sizeof(loc.lp_CountryData.cp_##what));

								LOCALESTRCOPY(DateTimeFormat);
								LOCALESTRCOPY(DateFormat);
								LOCALESTRCOPY(TimeFormat);
								LOCALESTRCOPY(ShortDateTimeFormat);
								LOCALESTRCOPY(ShortDateFormat);
								LOCALESTRCOPY(ShortTimeFormat);
								LOCALESTRCOPY(DecimalPoint);
								LOCALESTRCOPY(GroupSeparator);
								LOCALESTRCOPY(FracGroupSeparator);
								LOCALESTRCOPY(MonDecimalPoint);
								LOCALESTRCOPY(MonGroupSeparator);
								LOCALESTRCOPY(MonFracGroupSeparator);
								LOCALESTRCOPY(MonCS);
								LOCALESTRCOPY(MonSmallCS);
								LOCALESTRCOPY(MonIntCS);
								LOCALESTRCOPY(MonNegativeSign);
								LOCALESTRCOPY(MonPositiveSign);

								#define LOCALEBYTECOPY(what) \
									memcpy(loc.lp_CountryData.cp_##what,def->loc_##what,sizeof(loc.lp_CountryData.cp_##what));

								LOCALEBYTECOPY(Grouping);
								LOCALEBYTECOPY(FracGrouping);
								LOCALEBYTECOPY(MonGrouping);
								LOCALEBYTECOPY(MonFracGrouping);

								#define LOCALEVALEASSIGN(what) \
									loc.lp_CountryData.cp_##what = def->loc_##what;

								LOCALEVALEASSIGN(MonFracDigits);
								LOCALEVALEASSIGN(MonIntFracDigits);
								LOCALEVALEASSIGN(MonPositiveSpaceSep);
								LOCALEVALEASSIGN(MonPositiveSignPos);
								LOCALEVALEASSIGN(MonPositiveCSPos);
								LOCALEVALEASSIGN(MonNegativeSpaceSep);
								LOCALEVALEASSIGN(MonNegativeSignPos);
								LOCALEVALEASSIGN(MonNegativeCSPos);

								if ((ip = AllocIFF()) != NULL)
								{
									if ((ip->iff_Stream = Open("ENV:sys/locale.prefs",MODE_NEWFILE)))
									{
										InitIFFasDOS(ip);

										if (!OpenIFF(ip,IFFF_WRITE))
										{
											if (!PushChunk(ip, ID_PREF, ID_FORM, IFFSIZE_UNKNOWN))
											{
												if (!PushChunk(ip,ID_PREF,ID_PRHD,IFFSIZE_UNKNOWN))
												{
													struct PrefHeader prhd = {0};
													WriteChunkBytes(ip,&prhd,sizeof (prhd));
													PopChunk(ip);
												}

												if (!PushChunk(ip,ID_PREF,ID_LCLE,IFFSIZE_UNKNOWN))
												{
													WriteChunkBytes(ip,&loc,sizeof (loc));
													PopChunk(ip);
												}

												PopChunk(ip);
											}

											CloseIFF(ip);
										}

										Close(ip->iff_Stream);
									}

									FreeIFF(ip);
								}

								D(bug("Locale changed\n"));

								// check if we don't have to change the font too!
								i = 0;
								while (FontLanguages[i].Language)
								{
									if (!strcmp(FontLanguages[i].Language,loc.lp_PreferredLanguages[0]))
									{
										BPTR f;

										if ((f = Open("ENV:sys/font.conf",MODE_NEWFILE)) != NULL)
										{
											Write(f,FontLanguages[i].FontFile,strlen(FontLanguages[i].FontFile));
											Close(f);
										}
										break;
									}
									i++;
								}

								D(bug("Font changed\n"));
								CloseLocale(def);
							}
						}

						MUI_DisposeObject(app);
					}

					CloseLibrary(MUIMasterBase);
				}
			}

			DeletePool(pool);
		}
	}
}

/*********************************************************************************************/

WORD ShowMessage(LONG logo, STRPTR title, STRPTR text, STRPTR gadtext)
{
	struct EasyStruct es;

#warning "logo should be used to pick proper logo later on. - Piru"
	es.es_StructSize   = sizeof(es);
	es.es_Flags        = 0;
	es.es_Title        = title;
	es.es_TextFormat   = text;
	es.es_GadgetFormat = gadtext;

	return EasyRequestArgs(NULL, &es, NULL, NULL);
}

/*********************************************************************************************/

static void PrintMessage(LONG logo, STRPTR msg)
{
	if (msg)
	{
		if (IntuitionBase && !((struct Process *)FindTask(NULL))->pr_CLI)
		{
			ShowMessage(logo, "IPrefs", msg, "Ok");
		}
		else
		{
			Printf("IPrefs: %s\n", (LONG) msg);
		}
	}
}

/*********************************************************************************************/

static void PrintMessageReq(LONG logo, STRPTR msg)
{
	if (msg)
	{
		if (IntuitionBase)
		{
			ShowMessage(logo, "IPrefs", msg, "Ok");
		}
		else
		{
			Printf("IPrefs: %s\n", (LONG) msg);
		}
	}
}

/*********************************************************************************************/

static void	InitAC97Mixer(void)
{
	struct Library *AC97MixerBase;

	AC97MixerBase = OpenLibrary("ac97mixer.library", 0);
	if (AC97MixerBase)
	{
		CloseLibrary(AC97MixerBase);
	}
}

/*********************************************************************************************/

static ULONG	OpenLibs(void)
{
	struct libinfo *li;

	for (li = libtable; li->var; li++)
	{
		if (!((*(struct Library **)li->var) = OpenLibrary(li->name, li->version)))
		{
			sprintf(s, "Can't open %s V%ld!", li->name, (LONG) li->version);
			return(FALSE);
		}
	}
	return(TRUE);
}

/*********************************************************************************************/

static void CloseLibs(void)
{
	struct libinfo *li;

	for (li = libtable; li->var; li++)
	{
		if (!patches_installed || !li->nocloseafterpatch)
		{
			if (*(struct Library **)li->var)
			{
				CloseLibrary((*(struct Library **)li->var));
			}
		}
	}
}

/*********************************************************************************************/

static ULONG	GetENVName(void)
{
	BPTR lock;
	BOOL ok = FALSE;

	lock = Lock("ENV:", ACCESS_READ);
	if (lock)
	{
		strcpy(envname, "ENV:");
		ok = TRUE;

		UnLock(lock);
	}
	return(ok);
}

/*********************************************************************************************/

#define SYSNAME "sys"

static BPTR	CreateMissingDir(CONST_STRPTR name)
{
	BPTR lock;

	lock = Lock(name, ACCESS_READ);
	if (!lock && IoErr() == ERROR_OBJECT_NOT_FOUND)
	{
		lock = CreateDir(name);
		if (lock)
		{
			/* Get shared lock */
			UnLock(lock);
			lock = Lock(name, ACCESS_READ);
		}
	}

	return(lock);
}

static ULONG	FixENVDirs(CONST_STRPTR base)
{
	BOOL ok = FALSE;
	struct Process *me = (struct Process *) FindTask(NULL);
	APTR oldwinptr = me->pr_WindowPtr;
	BPTR baselock;

	me->pr_WindowPtr = (APTR) -1;

	baselock = Lock(base, ACCESS_READ);
	if (baselock)
	{
		BPTR olddir = CurrentDir(baselock);
		BPTR syslock;

		syslock = CreateMissingDir(SYSNAME);
		if (syslock)
		{
			BPTR netlock;

			/* Create xxx:sys/net */
			CurrentDir(syslock);
			netlock = CreateMissingDir("net");
			CurrentDir(baselock);
			if (netlock)
			{
				UnLock(netlock);
				ok = TRUE;
			}

			UnLock(syslock);
		}

		CurrentDir(olddir);

		UnLock(baselock);
	}

	me->pr_WindowPtr = oldwinptr;

	return(ok);
}

/*********************************************************************************************/

static ULONG	StartNotifications(void)
{
	WORD i;
	//BPTR lock;

	if ((notifyport = CreateMsgPort()))
	{
		notifysig = 1L << notifyport->mp_SigBit;

		for (i = 0; preftable[i].filename; i++)
		{
			if (!preftable[i].filenamebuffer)
			{
				continue;
			}

			strcpy(preftable[i].filenamebuffer, envname);
			AddPart(preftable[i].filenamebuffer, SYSNAME, 256);
			AddPart(preftable[i].filenamebuffer, preftable[i].filename, 256);

			/* If new prefs file exists, don't start notify for old */
			if (preftable[i].nextisfallback)
			{
				BPTR lock;

				D(bug("nextisfallback check\n"));

				lock = Lock(preftable[i].filenamebuffer, ACCESS_READ);
				if (lock)
				{
					D(bug("new prefs exists, stop next old\n"));
					UnLock(lock);
					preftable[i + 1].filenamebuffer = NULL;
				}
				else
				{
					D(bug("doesn't exist\n"));
				}
			}


#if 0
			lock = Lock(preftable[i].filenamebuffer,ACCESS_READ);
			if (!lock)
			{
				if (preftable[i].fallbackname)
				{
					strcpy(preftable[i].filenamebuffer, envname);
					AddPart(preftable[i].filenamebuffer, SYSNAME, 256);
					AddPart(preftable[i].filenamebuffer, preftable[i].fallbackname, 256);
					lock = Lock(preftable[i].filenamebuffer,ACCESS_READ);
				}
			}

			if (lock)
			{
				UnLock(lock);
			}
#endif

			preftable[i].nr.nr_Name     	    	= preftable[i].filenamebuffer;
			preftable[i].nr.nr_UserData 	    	= i;
			preftable[i].nr.nr_Flags    	    	= NRF_SEND_MESSAGE | NRF_NOTIFY_INITIAL;
			preftable[i].nr.nr_stuff.nr_Msg.nr_Port	= notifyport;

			D(bug("Trying to start notification for file <%s>.\n", preftable[i].filenamebuffer));

			if (StartNotify(&preftable[i].nr))
			{
				D(bug("Notification successfully started.\n"));

				preftable[i].notifystarted = TRUE;
			}
			else
			{
				D(bug("Notification start failed!! Continuing anyway!\n"));
			}
		} /* for (i = 0; preftable[i].filename; i++) */
	}
	return(TRUE);
}

/*********************************************************************************************/

static void KillNotifications(void)
{
	WORD i;

	for (i = 0; preftable[i].filename; i++)
	{
		if (preftable[i].notifystarted)
		{
			EndNotify(&preftable[i].nr);
			preftable[i].notifystarted = FALSE;
		}
	}

	if (notifyport)
	{
		DeleteMsgPort(notifyport);
		notifyport = NULL;
		notifysig = 0;
	}
}

/*********************************************************************************************/

static ULONG PreparePatches(void)
{
	struct IPrefsSem *sem;
	BOOL    	 created_sem = FALSE;

	if ((sem = AllocVec(sizeof(struct IPrefsSem),
	                    MEMF_PUBLIC | MEMF_CLEAR)))
	{
		InitSemaphore(&sem->sem);
		sem->sem.ss_Link.ln_Name = sem->semname;
		strcpy(sem->semname, IPREFS_SEM_NAME);

		Forbid();
		if (!(iprefssem = (struct IPrefsSem *)FindSemaphore(IPREFS_SEM_NAME)))
		{
			D(bug("IPrefs Semaphore 0x%p\n", iprefssem));

			iprefssem = sem;
			AddSemaphore(&iprefssem->sem);

			created_sem = TRUE;
		}
		Permit();

		if (created_sem)
		{
#ifndef __MORPHOS__
			InstallPatches();
#endif

			InstallDisplayBeepPatch();

			return TRUE;
		}
		else
		{
			FreeVec(sem);
		}
	}
	return FALSE;
}

/*********************************************************************************************/

void ParsePrefsFile(STRPTR filename)
{
	struct IFFHandle *iff;

	D(bug("IPrefsHandler: Parse <%s>\n", filename));

	//if ((IFFParseBase = OpenLibrary("iffparse.library", 0)))
	{
		if ((iff = AllocIFF()))
		{
			if ((iff->iff_Stream = Open(filename, MODE_OLDFILE)))
			{
				InitIFFasDOS(iff);

				if (OpenIFF(iff, IFFF_READ) == 0 &&
				    ParseIFF(iff, IFFPARSE_STEP) == 0)
				{
					struct ContextNode *cn = CurrentChunk(iff);

					D(bug("iff %p cn %p Type %08lx ID %08lx\n", iff, cn, cn ? cn->cn_Type : 0, cn ? cn->cn_ID : 0));
					if (cn && cn->cn_Type == ID_PREF && cn->cn_ID == ID_FORM)
					{
						int k;
						LONG error;

						D(bug("ID_PREF\n"));

						for (k = 0; k < sizeof(scantable) / sizeof(scantable[0]); ++k)
						{
							error = EntryHandler(iff, ID_PREF, scantable[k].id, IFFSLI_ROOT, &scantable[k].hook, iff);
							D(bug("IPrefsHandler: Handler %d (%08lx) error %d\n", k, scantable[k].id, error));
						}

						do
						{
							error = ParseIFF(iff, IFFPARSE_SCAN);
							D(bug("IPrefsHandler: OpenIFF/ParseIFF failed, error %d\n",error));
						}
						while (error == 0 || error == IFFERR_EOC);
					}
					CloseIFF(iff);
				}
				else
				{
					D(bug("IPrefsHandler: OpenIFF/ParseIFF failed\n"));
				}
				D(bug("IPrefsHandler: Close iff_Stream 0x%lx\n",iff->iff_Stream));
				Close(iff->iff_Stream);
			}
			else
			{
				D(bug("IPrefsHandler: file not found\n"));
			}
			FreeIFF(iff);
		}
		//CloseLibrary(IFFParseBase);
	}

	/* Handle WB 3.5/3.9 HiddenDevice entries - Piru */

	if (!IsListEmpty((struct List *) &HiddenDeviceList))
	{
		struct Node *Node;

		if (WorkbenchBase->lib_Version >= 44)
		{
			struct List *hdlist = NULL;

			if (WorkbenchControl(NULL,
			                     WBCTRLA_GetHiddenDeviceList, (ULONG) &hdlist,
			                     TAG_DONE) &&
			    hdlist)
			{
				ForeachNode(hdlist, Node)
				{
					if (Node->ln_Type == 0x67)
					{
						WorkbenchControl(NULL,
						                 WBA_Reserved9, (ULONG) Node->ln_Name,
						                 TAG_DONE);
					}
				}

				WorkbenchControl(NULL,
				                 WBCTRLA_FreeHiddenDeviceList, (ULONG) hdlist,
				                 TAG_DONE);
			}
		}

		while ((Node = REMHEAD((struct List *) &HiddenDeviceList)))
		{
			if (WorkbenchBase->lib_Version >= 44)
			{
				WorkbenchControl(NULL,
				                 WBA_Reserved8, (ULONG) Node->ln_Name,
				                 TAG_DONE);
			}

			FreeVec(Node);
		}

		if (WorkbenchBase->lib_Version >= 44)
		{
			WorkbenchControl(NULL,
			                 WBA_Reserved10, TRUE,
			                 TAG_DONE);
		}
	}
}

/*********************************************************************************************/

#if USE_DELAYEDREOPENWB

static struct MsgPort		*ReOpenWBTimerPort = NULL;
struct timerequest		*ReOpenWBtimereq = NULL;
BOOL				ReOpenWBtimer;

static BOOL StartReOpenWB(void)
{
	ReOpenWBTimerPort = CreateMsgPort();
	if (ReOpenWBTimerPort)
	{
		ReOpenWBtimereq = (struct timerequest *) CreateIORequest(ReOpenWBTimerPort, sizeof(struct timerequest));
		if (ReOpenWBtimereq)
		{
			if (OpenDevice(TIMERNAME, UNIT_VBLANK, (struct IORequest *) ReOpenWBtimereq, 0) == 0)
			{
				reopenwbtimersig = 1L << ReOpenWBTimerPort->mp_SigBit;
				ReOpenWBtimer = FALSE;

				return TRUE;
			}

			DeleteIORequest((struct IORequest *) ReOpenWBtimereq);
		}

		DeleteMsgPort(ReOpenWBTimerPort);
	}

	return FALSE;
}

static void CleanupReOpenWB(void)
{
	if (ReOpenWBTimerPort)
	{
		if (ReOpenWBtimereq)
		{
			if (ReOpenWBtimer)
			{
				AbortIO((struct IORequest *) ReOpenWBtimereq);
				WaitIO((struct IORequest *) ReOpenWBtimereq);

				ReOpenWBtimer = FALSE;
			}

			CloseDevice((struct IORequest *) ReOpenWBtimereq);
			DeleteIORequest((struct IORequest *) ReOpenWBtimereq);

			ReOpenWBtimereq = NULL;
		}

		DeleteMsgPort(ReOpenWBTimerPort);

		ReOpenWBTimerPort = NULL;
		reopenwbtimersig = 0;
	}
}

#else /* USE_DELAYEDREOPENWB */

static BOOL StartReOpenWB(void)
{
	return TRUE;
}

static void CleanupReOpenWB(void)
{
}

#endif /* USE_DELAYEDREOPENWB */


/*********************************************************************************************/

static LONG handlesigs(ULONG sigs);

static void IPrefsHandler(void)
{
	LONG gmsgreplied = FALSE;

	D(bug("IPrefsHandler:\n"));

	InitAC97Mixer();

	HandlerProcess = (struct Process *) FindTask(NULL);
	NEWLIST((struct List *) &HiddenDeviceList);
	PrevLockPens = FALSE;

	if (StartReOpenWB())
	{
		if (StartNotifications())
		{
			if (PreparePatches())
			{
				/* C= IPrefs uses following signals:
				 * CTRL_C - notify signal
				 * CTRL_D - audio.devide ioreq signal
				 * CTRL_E - wbpattern reload notify
				 * CTRL_F - displaybeep notify
				 *
				 * We use:
				 * CTRL_C - exit, not really supported
				 * CTRL_D - wbpattern reload notify
				 * CTRL_E - displaybeep notify
				 */

				ULONG waitmask = notifysig | reopenwbtimersig | SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D | SIGBREAKF_CTRL_E;
				ULONG sigs;
				LONG  quit = FALSE;

				/*
				 * This delay supposedly gives filesystem time to trigger
				 * notification messages. IMO it's not needed since
				 * NRF_NOTIFY_INITIAL is immediate (triggered from within
				 * ACTION_ADD_NOTIFY that is DoPkt()'d by dos StartNotify).
				 * - Piru
				 */
				//Delay(5);

				/*
				 * Start Network when no config file was found which
				 * would trigger the notify prefs load/startup
				 */
				StartNetwork();

				#if USE_NEWSTARTUPSYNC
				D(bug("IPrefsHandler: Handle initial Prefs Notification\n"));

				/*
				 * Poll signals once to handle prefs load before gmsg reply.
				 * This (supposedly) allows prefs to kick in before iprefs
				 * lets system continue booting. Without such syncronization
				 * Ambient would first open in 640x480 and then reopen in
				 * saved screenmode, fucking everything (icon positions)
				 * up... - Piru
				 */
				sigs = SetSignal(0, waitmask) & waitmask;
				if (sigs)
				{
					quit = handlesigs(sigs);
				}

				D(DumpFonts());

				/* If this is a CD Boot, ask the user to pick his preferred language/keymap layout */
				SetBootCDLanguageKB();

				/*
				 * Poll signals again before continuing...
				*/
				sigs = SetSignal(0, waitmask) & waitmask;
				if (sigs)
				{
					quit = handlesigs(sigs);
				}
				
				D(DumpFonts());
				
				/* We can continue main app */
				gmsgreplied = TRUE;
				PutMsg(grport, &gmsg);

				#endif /* USE_NEWSTARTUPSYNC */


				D(bug("IPrefsHandler: Wait for Prefs Notification\n"));

				while (!quit)
				{
					sigs = Wait(waitmask | GetWaitSampleMask());

					D(bug("IPrefsHandler: SigMask 0x%lx\n", sigs));

					quit = handlesigs(sigs);

					#if !USE_NEWSTARTUPSYNC
					if (sigs & notifysig)
					{
						/*
						 * All this only works *IF* notification(entry) is done instantly
						 * which we can assume because the iprefs task runs at pri 0 and
						 * filesystems usually at 10
						 */
						if (!gmsgreplied)
						{
							gmsgreplied = TRUE;
							PutMsg(grport, &gmsg);
						}
					}
					#endif /* !USE_NEWSTARTUPSYNC */

				} /* while (!quit) */

				/* Quit is really not supported - patches are not removed */
			}
			else
			{
				PrintMessage(-1, "Can't install patches!");
			}

			KillNotifications();
		}
		else
		{
			PrintMessage(-1, "Can't create notification msg port!");
		}

		CleanupReOpenWB();
	}
	else
	{
		PrintMessage(-1, "Can't create timer msg port!");
	}
	D(bug("IPrefsHandler: finished\n"));

	#if 1
	/* Let the main app continue if not yet done */
	if (!gmsgreplied)
	{
		PutMsg(grport, &gmsg);
	}
	#endif
}

/*********************************************************************************************/

static LONG handlesigs(ULONG sigs)
{
	if (sigs & SIGBREAKF_CTRL_E)
	{
		D(bug("IPrefsHandler: DisplayBeep\n"));
		PlayBeep();
	}

	if (sigs & GetWaitSampleMask())
	{
		D(bug("IPrefsHandler: CleanupSample()\n"));
		CleanupSample();
	}

	if (sigs & SIGBREAKF_CTRL_D)
	{
		D(bug("IPrefsHandler: LoadPatterns\n"));
		LoadPatterns();
	}

#if USE_DELAYEDREOPENWB
	if ((sigs & reopenwbtimersig) && ReOpenWBtimer)
	{
		DoReOpenWBScreen();
		ReOpenWBtimer = FALSE;
	}
#endif /* USE_DELAYEDREOPENWB */

	if (sigs & notifysig)
	{
		struct NotifyMessage *msg;

		D(bug("IPrefsHandler: Notification\n"));
		while((msg = (struct NotifyMessage *)GetMsg(notifyport)))
		{
			WORD id = msg->nm_NReq->nr_UserData;

			D(bug("IPrefsHandler: Received notify message. UserData = %d --> File = \"%s\"\n", id,
			      preftable[id].filenamebuffer));

			if (preftable[id].handler)
			{
				(*preftable[id].handler)(preftable[id].filenamebuffer);
			}
			else
			{
				ParsePrefsFile(preftable[id].filenamebuffer);
			}

			ReplyMsg(&msg->nm_ExecMessage);

		} /* while((msg = (struct NotifyMessage *)GetMsg(notifyport))) */
	} /* if (sigs & notifysig) */

	if (sigs & SIGBREAKF_CTRL_C)
	{
		D(bug("IPrefsHandler: Break\n"));
		return TRUE;
	}

	return FALSE;
}

/*********************************************************************************************/

static void DumpFonts()
{
	struct TextFont *f;
	
	Forbid();
	
	dprintf("---Begin font dump\n");
	
	ForeachNode(&GfxBase->TextFonts, f)
	{
		dprintf("Font %s\n", f->tf_Message.mn_Node.ln_Name);
	}
	
	dprintf("---End font dump\n");

	Permit();
}
